//------------------------------------------------------------------------------
//
//  npdata.c
//
// ANTZ - 3D data viewer with realtime real-world interaction based on NPE.
//
// ANTZ is hosted at http://temporalzone.com and NPE at http://neuralphysics.org
//
// Written in 2010-2011 by Shane Saxon - contact@saxondigital.net
//
// Additional code contributed by: Mark Buchanan and Lucas Erickson.
//
// To the extent possible under law, the author(s) have dedicated all copyright 
// and related and neighboring rights to this software to the public domain
// worldwide. This software is distributed without any warranty.
//
// Released under the CC0 license, which is GPL compatible.
//
// You should have received a copy of the CC0 Public Domain Dedication along
// with this software (license file named COPYING.txt). If not, see
// http://creativecommons.org/publicdomain/zero/1.0/
//
//------------------------------------------------------------------------------


#include "npdata.h"
#include "npctrl.h"
#include "data/npmap.h"
#include "io/npkey.h"
#include "io/npfile.h"
#include "io/npmouse.h"


void* gData;

void npInitDataMap (void* dataRef);
void npInitDataIO (int argc, char** argv, void* dataRef);
void npInitDataCtrl (void* dataRef);


//------------------------------------------------------------------------------
void* npInitData (int argc, char** argv)
{
	// allocate memory for the global 'data' structure
	pData data = (pData) malloc (sizeof(Data));
	if (data == NULL)
	{
		printf ("err 4213 - malloc failed cannot run app \n");
		return NULL;
	}

	printf ("Starting...\n");

	// initialize data
	npInitDataMap (data);
	npInitDataIO (argc, argv, data);	//stores the command line arguements
	npInitDataCtrl (data);


	gData = data;						//store a copy for retrieval

	return gData;
	// create a fixed array of global data pointers that is thread safe, zz
	// need to support multiple instances within a single app
}


//------------------------------------------------------------------------------
void npInitDataMap(void* dataRef)
{
	int i = 0;
	pData data = (pData) dataRef;

	NPnodePtr node = NULL;
	NPnodePtr nodeChild = NULL;

	npInitMap (dataRef);
	npInitNode (dataRef);

	// global root reserved at index = 0
	npNodeNew (kNodeDefault, NULL, dataRef);

	// root camera, reserved at index = 1
	node = npNodeNew (kNodeCamera, NULL, dataRef);
	npDataCameraPreset (0, node, dataRef);

	// child cameras attached to root camera
	for (i=0; i < kNPcameraCount; i++)
	{
		nodeChild = npNodeNew (kNodeCamera, node, dataRef);
		npDataCameraPreset (i, nodeChild, dataRef);
	}
	// select the first child camera
	npCtrlCommand (kNPgraphCamera, dataRef);

	// root grid at index = 2
	node = npNodeNew (kNodeGrid, NULL, dataRef);
//	node->scale.z = 1.5f;
	node->textureID = 1;		// first texture is defualt for the grid

	// setup initial scene using preset 1
	//	npDataPreset (1, dataRef);

	// re-select the camera
	npCtrlCommand (kNPgraphCamera, dataRef);
}

//------------------------------------------------------------------------------
void npInitDataGL(void* dataRef)
{
	pData data = (pData) dataRef;
	
	data->io.gl.normal = kNPglNormalize;
	data->io.gl.shade = kNPglShadeSmooth; //kNPglShadeFlat; //works if starts Flat, debug zz
	
	data->io.gl.alphaMode = kNPalphaModeSubtractive;

	data->io.gl.textureCount = 0;

	data->io.gl.pickPass = false;
	data->io.gl.pickMode = kNPpickModePin;
	data->io.gl.pickID = 0;

	data->io.gl.width = 0;
	data->io.gl.height = 0;
}

//------------------------------------------------------------------------------
void npInitDataIO(int argc, char** argv, void* dataRef)
{
	int i = 0, j = 0;

	pData data = (pData) dataRef;

	// setup initial key command assignments
	npInitKey (dataRef);

	npInitDataGL (dataRef);

	// store reference to the command line arguements
	data->io.argc = argc;
	data->io.argv = argv;


	data->io.drawMenu = false;

	data->io.clearFrameBuffer = true;
	data->io.clearIndex = 0;

	data->io.clear.r = 0.0f;
	data->io.clear.g = 0.0f;
	data->io.clear.b = 0.0f;
	data->io.clear.a = 0.0f;

	data->io.axes.x = true;
	data->io.axes.y = true;
	data->io.axes.z = true;


	for ( i=0; i < kNPchannelMax; i++)
		for ( j=0; j < kNPchannelSize; j++)
			data->io.channel[i][j] = 0.0f;
	
	data->io.channelCount = 30;
	data->io.channelIndex = 0;
	data->io.channelOldest = 0;
	data->io.channelSize = 0;


	// clear arrays
	for (i=0; i < kNPdataTypeMax; i++)
		data->io.write[i] = NULL;

	// setup the default URL
	for (i=0; i < kNPurlMax; i++)
		data->io.url[i] = 0;
	sprintf (data->io.url, "http://temporalzone.com/id.html");

	data->io.size = sizeof(NPio);
}


//------------------------------------------------------------------------------
void npInitDataCtrl(void* dataRef)
{
	pData data = (pData) dataRef;

	data->ctrl.slow = 0.2f;		//normalize speed to 1.0 and 5.0, debug zz
	data->ctrl.fast = 2.0f;		//do this in npengine

	data->ctrl.size = sizeof(NPctrl);
}


//------------------------------------------------------------------------------
void npCloseData()
{
	pData data = (pData) gData;
	
	free (data);
}


//------------------------------------------------------------------------------
void* npGetDataRef()
{
	return gData;
}


//------------------------------------------------------------------------------
void npUpdateData (void* dataRef)
{
	pData data = (pData) dataRef;

	//update the data->map

	return;
}

//------------------------------------------------------------------------------
char* npGetDataTypeName (int type)
{
	if (type > 0 )//&& type < kTypeCount)
		return NULL;//(data->map.typeList[type].name);
	else
		return NULL;
}
//------------------------------------------------------------------------------
char* npGetDataTypeDesc (int type)
{
	if (type > 0 )//&& type < kTypeCount)
		return NULL;//(data->map.typeList[type].description);
	else
		return NULL;
}


// loads preset scene using currently selected object as the origin
// if alt is pressed (ie ALT+5) it loads preset at the global origin
//------------------------------------------------------------------------------
void npDataPreset (int preset, void* dataRef)
{
	int i = 0;
	int j = 0;

	NPfloatXYZ	translateOffset;
	NPfloatXYZ	rotateOffset;
	NPfloatXYZ	scaleOffset;

	pData data = (pData) dataRef;
	NPnodePtr node = data->map.node[data->map.nodeRootIndex];

	NPnodePtr nodeChild = NULL;

	NPpinPtr pinData = NULL;

	// don't offset if alt key held or camera selected
	if (data->io.key.modAlt || data->map.nodeRootIndex == kNPnodeRootPin)
	{
		translateOffset.x = 0.0f;
		translateOffset.y = 0.0f;
		translateOffset.z = 0.0f;

		rotateOffset.x = 0.0f;
		rotateOffset.y = 0.0f;
		rotateOffset.z = 0.0f;

		scaleOffset.x = 0.0f;
		scaleOffset.y = 0.0f;
		scaleOffset.z = 0.0f;
	}
	else
	{
					// add code to calc translate, rotate position based on
					// global coordinates think of GL order of operation,	debug zz
		translateOffset.x = node->translate.x;
		translateOffset.y = node->translate.y;
		translateOffset.z = node->translate.z;

		rotateOffset.x = node->rotate.x;
		rotateOffset.y = node->rotate.y;
		rotateOffset.z = node->rotate.z;

		scaleOffset.x = node->scale.x;
		scaleOffset.y = node->scale.y;
		scaleOffset.z = node->scale.z;
	}

	switch (preset)
	{
		case kNPgraphPresetOne :

			node = npNodeNew (kNodePin, NULL, dataRef);
				node->rotate.z = 1.7f;
				node->scale.x = 1.5f;
				node->scale.y = 1.5f;
				node->scale.z = 1.5f;

				node->translate.x -= 3.0f;
				node->translate.y -= 22.0f;

				node->colorIndex = 0;

			//create primary torus
			node = nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;

			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 0;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 0;

			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 1;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 4;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 3;
								node->rotate.x = -0.7f;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 3;
								node->rotate.x = 0.7f;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 4;
								node->rotate.x = 0.7f;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 2;
								node->rotate.x = 0.7f;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 1;
								node->rotate.x = 1.5f;

			npCtrlCommand (kNPgraphSelectAll, data);
			npCtrlCommand (kNPgraphColorUp, data);
			npCtrlCommand (kNPgraphColorDown, data);
			npCtrlCommand (kNPgraphSelectAll, data);

			data->map.nodeRootIndex = data->map.nodeRootCount - 1;
			data->map.selectedIndex = data->map.nodeRootIndex;
			data->map.selectedPinNode = data->map.node[data->map.selectedIndex];
			data->map.activeNode = data->map.selectedPinNode;
	
			break;
		case kNPgraphPresetTwo :
			// draw 6x6 grid of pins
			for(i = -3; i < 3; i++)
				for(j=-3; j < 3; j++)
				{
					// create the root pin
					node = npNodeNew (kNodePin, NULL, dataRef);
					node->translate.x = i * 10.0f;
					node->translate.y = j * 10.0f;

					node->translate.x += translateOffset.x;
					node->translate.y += translateOffset.y;
					node->translate.z += translateOffset.z;

						node->colorIndex = 0;
					node = nodeChild = npNodeNew (kNodePin, node, dataRef);
						nodeChild->colorIndex = 14;

					nodeChild = npNodeNew (kNodePin, node, dataRef);
						nodeChild->colorIndex = 1;
					nodeChild = npNodeNew (kNodePin, node, dataRef);
						nodeChild->colorIndex = 14;
					nodeChild = npNodeNew (kNodePin, node, dataRef);
						nodeChild->colorIndex = 1;
					node = npNodeNew (kNodePin, node, dataRef);
						node->colorIndex = 2;
					nodeChild = npNodeNew (kNodePin, node, dataRef);
						nodeChild->colorIndex = 1;
					node = npNodeNew (kNodePin, node, dataRef);
						node->colorIndex = 3;
					nodeChild = npNodeNew (kNodePin, node, dataRef);
						nodeChild->colorIndex = 4;
					data->map.nodeRootIndex = data->map.nodeRootCount - 1;
					data->map.selectedIndex = data->map.nodeRootIndex;
					data->map.selectedPinNode = data->map.node[data->map.selectedIndex];
					data->map.activeNode = data->map.selectedPinNode;
				}			
			break;
		case kNPgraphPresetThree :
			//first pin root
			node = npNodeNew (kNodePin, NULL, dataRef);
									node->rotateRate.z = 0.001f;
				node->colorIndex = 0;
			node = nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;
			//first pin children
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 2;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 3;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 4;

			//next pin
			node = npNodeNew (kNodePin, NULL, dataRef);
									node->rotateRate.z = -0.001f;
				node->colorIndex = 0;
			node = nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;

			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 2;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 3;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 4;

			//next pin
			node = npNodeNew (kNodePin, NULL, dataRef);
				node->colorIndex = 0;
			node = nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 0;

			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			node = npNodeNew (kNodePin, node, dataRef);
								node->rotateRate.x = 0.001f;
				node->colorIndex = 2;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			node = npNodeNew (kNodePin, node, dataRef);
									node->rotateRate.x = 0.005f;
				node->colorIndex = 3;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 4;
			
			//next pin
			node = npNodeNew (kNodePin, NULL, dataRef);
									node->rotateRate.z = 0.005f;
				node->colorIndex = 0;
			node = nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;

			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 2;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 2;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 3;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 4;

			//next pin
			node = npNodeNew (kNodePin, NULL, dataRef);
									node->rotateRate.z = 0.001f;
				node->colorIndex = 0;
			node = nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;

			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 2;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 3;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 4;

			npCtrlCommand (kNPgraphSelectAll, data);
			npCtrlCommand (kNPgraphColorUp, data);
			npCtrlCommand (kNPgraphColorDown, data);
			npCtrlCommand (kNPgraphSelectAll, data);

				
			data->map.nodeRootIndex = data->map.nodeRootCount - 1;
			data->map.selectedIndex = data->map.nodeRootIndex;
			data->map.selectedPinNode = data->map.node[data->map.selectedIndex];
			data->map.activeNode = data->map.selectedPinNode;

			npCtrlCommand (kNPgraphNext, data);
			npCtrlCommand (kNPgraphNext, data);
			npCtrlCommand (kNPgraphNext, data);
			npCtrlCommand (kNPgraphNext, data);

			break;
		case kNPgraphPresetFour : 
			
			node = npNodeNew (kNodePin, NULL, dataRef);
				node->colorIndex = 0;

			// primary torus, branchLevel == 1 
			node = npNodeNew (kNodePin, node, dataRef);//node->child[0];
				node->colorIndex = 14;
			// child branchLevel == 2
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 14;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;

			// select child, branchLevel == 3
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 2;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 1;

			// next branchLevel
			node = npNodeNew (kNodePin, node, dataRef);
				node->colorIndex = 3;
			nodeChild = npNodeNew (kNodePin, node, dataRef);
				nodeChild->colorIndex = 4;

			npCtrlCommand (kNPgraphSelectAll, data);
			npCtrlCommand (kNPgraphColorUp, data);
			npCtrlCommand (kNPgraphColorDown, data);
			npCtrlCommand (kNPgraphSelectAll, data);

			data->map.nodeRootIndex = data->map.nodeRootCount - 1;
			data->map.selectedIndex = data->map.nodeRootIndex;
			data->map.selectedPinNode = data->map.node[data->map.selectedIndex];
			data->map.activeNode = data->map.selectedPinNode;

			break;
		case kNPgraphPresetFive : break;
		case kNPgraphPresetSix : break;
		case kNPgraphPresetSeven : break;
		case kNPgraphPresetEight : break;

		default : break;
	}

	npCtrlCommand (kNPgraphCamera, data);			// select the camera
}


//resets the camera node passed in, specific to the preset number
//------------------------------------------------------------------------------
void npDataCameraPreset (int preset, NPnodePtr node, void* dataRef)
{
	int id = 0;	
	int branchLevel = 0;

	NPnodePtr nodeParent = NULL;
	
	// retain parameters
//	id = node->id;
//	branchLevel = node->branchLevel;
//	nodeParent = node->parent;

	//reset to defaults
//	npInitNodeDefault (node);			//need to retain most values, debug zz
	npInitNodeCamera (node);

	//restore 
//	node->id = id;
//	node->branchLevel = branchLevel;
//	node->parent = nodeParent;

	//set preset values
	switch (preset)
	{
		case 0 :
			node->translate.x = -27.0f;
			node->translate.y = -25.0f;
			node->translate.z = 9.0f;
			
			node->rotateVec.angle = 0.0f;
			node->rotateVec.x = 0.0f;
			node->rotateVec.y = 1.0f;
			node->rotateVec.z = 0.0f;

			node->rotate.x = kPos2PI / 32.0f;
			node->rotate.z = kPos2PI / 12.0f;
			break;
		case 1 : 
			node->translate.x = -0.0f;
			node->translate.y = -0.0f;
			node->translate.z = 100.0f;
			
			node->rotateVec.angle = 0.0f;
			node->rotateVec.x = 0.0f;
			node->rotateVec.y = 1.0f;
			node->rotateVec.z = 0.0f;

			node->rotate.x = kPos2PI / 4.0f;
			break;
		case 2 : 
			node->translate.x = -0.0f;
			node->translate.y = -85.0f;
			node->translate.z = 7.0f;
			
			node->rotateVec.angle = 0.0f;
			node->rotateVec.x = 0.0f;
			node->rotateVec.y = 1.0f;
			node->rotateVec.z = 0.0f;
			break;
		case 3 : 
			node->translate.x = 85.0f;
			node->translate.y = 0.0f;
			node->translate.z = 7.0f;
			
			node->rotateVec.angle = 0.0f;
			node->rotateVec.x = 0.0f;
			node->rotateVec.y = 1.0f;
			node->rotateVec.z = 0.0f;

			node->rotate.z = -kPos2PI / 4.0f;		
			break;
		default : 
			node->translate.x = -27.0f;
			node->translate.y = -25.0f;
			node->translate.z = 9.0f;
			
			node->rotateVec.angle = 0.0f;
			node->rotateVec.x = 0.0f;
			node->rotateVec.y = 1.0f;
			node->rotateVec.z = 0.0f;

			node->rotate.x = kPos2PI / 32.0f;
			node->rotate.z = kPos2PI / 12.0f;
			//	node->triggerOnLo.z = true;		//prevents going below surface
			//	node->triggerLo.z = 1.5f;
			break;
	}
}

